// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_lib_interpolation_scalar.hpp"
#include "mo_lib_loopindices.hpp"
#include <Kokkos_Core.hpp>
#include <iostream>

//-----------------------------------------------------------------------
//
//  ! averaging and interpolation routines and
//  ! routines needed to compute the coefficients therein
//
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
//>
///  Performs  average of scalar fields from vertices to velocity points.
///
///  The coefficients are given by coeff_int.
///
template <typename T>
void verts2edges_scalar_lib(const T *p_vertex_in, const int *edge_vertex_idx,
                            const int *edge_vertex_blk, const T *coeff_int,
                            T *p_edge_out, const int i_startblk,
                            const int i_endblk, const int i_startidx_in,
                            const int i_endidx_in, const int slev,
                            const int elev, const int nproma, const int nlev,
                            const int nblks_v, const int nblks_e,
                            const bool lacc) {

  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  UnmanagedConstT3D p_vertex_in_view(p_vertex_in, nproma, nlev, nblks_v);
  UnmanagedConstInt3D iidx_view(edge_vertex_idx, nproma, nblks_e, 4);
  UnmanagedConstInt3D iblk_view(edge_vertex_blk, nproma, nblks_e, 4);
  UnmanagedConstT3D coeff_int_view(coeff_int, nproma, 2, nblks_e);
  UnmanagedT3D p_edge_out_view(p_edge_out, nproma, nlev, nblks_e);

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

    int i_startidx, i_endidx;
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "verts2edges_scalar", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int je) {
          p_edge_out_view(je, jk, jb) =
              coeff_int_view(je, 0, jb) *
                  p_vertex_in_view(iidx_view(je, jb, 0), jk,
                                   iblk_view(je, jb, 0)) +
              coeff_int_view(je, 1, jb) *
                  p_vertex_in_view(iidx_view(je, jb, 1), jk,
                                   iblk_view(je, jb, 1));
        });
    Kokkos::fence();
  }
}

//------------------------------------------------------------------------
//>
///  Computes  average of scalar fields from centers of triangular faces to.
///
///  Computes  average of scalar fields from centers of triangular faces to
///  velocity points.
///
template <typename T, typename S>
void cells2edges_scalar_lib(const T *p_cell_in, const int *edge_cell_idx,
                            const int *edge_cell_blk, const S *coeff_int,
                            S *p_edge_out, const int *i_startblk_in,
                            const int *i_endblk_in, const int *i_startidx_in,
                            const int *i_endidx_in, const int slev,
                            const int elev, const int nproma, const int nlev,
                            const int nblk_c, const int nblks_e,
                            const int patch_id, const bool l_limited_area,
                            const bool lfill_latbc, const bool lacc) {

  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<const S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstS3D;
  typedef Kokkos::View<S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedS3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  UnmanagedConstT3D p_cell_in_view(p_cell_in, nproma, nlev, nblk_c);
  UnmanagedConstInt3D iidx_view(edge_cell_idx, nproma, nblks_e, 2);
  UnmanagedConstInt3D iblk_view(edge_cell_blk, nproma, nblks_e, 2);
  UnmanagedConstS3D coeff_int_view(coeff_int, nproma, 2, nblks_e);
  UnmanagedS3D p_edge_out_view(p_edge_out, nproma, nlev, nblks_e);

  // Fill outermost nest boundary
  int i_startblk, i_endblk;
  if ((l_limited_area || patch_id > 0) && (lfill_latbc)) {
    i_startblk = i_startblk_in[0];
    i_endblk = i_endblk_in[0];

    for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

      int i_startidx, i_endidx;
      get_indices_e_lib(i_startidx_in[0], i_endidx_in[0], nproma, jb,
                        i_startblk, i_endblk, i_startidx, i_endidx);

      Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
          {slev, i_startidx}, {elev + 1, i_endidx + 1});

      Kokkos::parallel_for(
          "cells2edges_scalar", innerPolicy,
          KOKKOS_LAMBDA(const int jk, const int je) {
            if (iidx_view(je, jb, 0) >= 0 && iblk_view(je, jb, 0) >= 0) {
              p_edge_out_view(je, jk, jb) = p_cell_in_view(
                  iidx_view(je, jb, 0), jk, iblk_view(je, jb, 0));
            } else if (iidx_view(je, jb, 1) >= 0 && iblk_view(je, jb, 1) >= 0) {
              p_edge_out_view(je, jk, jb) = p_cell_in_view(
                  iidx_view(je, jb, 1), jk, iblk_view(je, jb, 1));
            } else {
              std::cerr << "mo_interpolation:cells2edges_scalar_lib: error in "
                           "lateral boundary filling"
                        << std::endl;
              std::exit(EXIT_FAILURE);
            }
          });
      Kokkos::fence();
    }
  } else {
    // Process the remaining grid points for which a real interpolation is
    // possible
    i_startblk = i_startblk_in[1];
    i_endblk = i_endblk_in[1];

    for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

      int i_startidx, i_endidx;
      get_indices_e_lib(i_startidx_in[1], i_endidx_in[1], nproma, jb,
                        i_startblk, i_endblk, i_startidx, i_endidx);

      Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
          {slev, i_startidx}, {elev + 1, i_endidx + 1});

      Kokkos::parallel_for(
          "cells2edges_scalar", innerPolicy,
          KOKKOS_LAMBDA(const int jk, const int je) {
            p_edge_out_view(je, jk, jb) =
                coeff_int_view(je, 0, jb) *
                    p_cell_in_view(iidx_view(je, jb, 0), jk,
                                   iblk_view(je, jb, 0)) +
                coeff_int_view(je, 1, jb) *
                    p_cell_in_view(iidx_view(je, jb, 1), jk,
                                   iblk_view(je, jb, 1));
          });
      Kokkos::fence();
    }
  }
}

//------------------------------------------------------------------------
//>
///  Computes average of scalar fields from velocity points to.
///
///  Computes average of scalar fields from velocity points to
///  centers of dual faces.
///
template <typename T>
void edges2verts_scalar_lib(const T *p_edge_in, const int *vert_edge_idx,
                            const int *vert_edge_blk, const T *v_int,
                            T *p_vert_out, const int i_startblk,
                            const int i_endblk, const int i_startidx_in,
                            const int i_endidx_in, const int slev,
                            const int elev, const int nproma, const int nlev,
                            const int nblks_e, const int nblks_v,
                            const bool lacc) {

  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  UnmanagedConstT3D p_edge_in_view(p_edge_in, nproma, nlev, nblks_e);
  UnmanagedConstInt3D iidx_view(vert_edge_idx, nproma, nblks_v, 6);
  UnmanagedConstInt3D iblk_view(vert_edge_blk, nproma, nblks_v, 6);
  UnmanagedConstT3D v_int_view(v_int, nproma, 6, nblks_v);
  UnmanagedT3D p_vert_out_view(p_vert_out, nproma, nlev, nblks_v);

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

    int i_startidx, i_endidx;
    get_indices_v_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "edges2verts_scalar", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jv) {
          p_vert_out_view(jv, jk, jb) =
              v_int_view(jv, 0, jb) * p_edge_in_view(iidx_view(jv, jb, 0), jk,
                                                     iblk_view(jv, jb, 0)) +
              v_int_view(jv, 1, jb) * p_edge_in_view(iidx_view(jv, jb, 1), jk,
                                                     iblk_view(jv, jb, 1)) +
              v_int_view(jv, 2, jb) * p_edge_in_view(iidx_view(jv, jb, 2), jk,
                                                     iblk_view(jv, jb, 2)) +
              v_int_view(jv, 3, jb) * p_edge_in_view(iidx_view(jv, jb, 3), jk,
                                                     iblk_view(jv, jb, 3)) +
              v_int_view(jv, 4, jb) * p_edge_in_view(iidx_view(jv, jb, 4), jk,
                                                     iblk_view(jv, jb, 4)) +
              v_int_view(jv, 5, jb) * p_edge_in_view(iidx_view(jv, jb, 5), jk,
                                                     iblk_view(jv, jb, 5));
        });
    Kokkos::fence();
  }
}

//------------------------------------------------------------------------
//>
///  Computes interpolation from edges to cells
///
///  Computes interpolation of scalar fields from velocity points to
///  cell centers via given interpolation weights
///
template <typename T>
void edges2cells_scalar_lib(const T *p_edge_in, const int *edge_idx,
                            const int *edge_blk, const T *coeff_int,
                            T *p_cell_out, const int i_startblk,
                            const int i_endblk, const int i_startidx_in,
                            const int i_endidx_in, const int slev,
                            const int elev, const int nproma, const int nlev,
                            const int nblks_e, const int nblks_c,
                            const bool lacc) {
  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  // edge based scalar input field, dim: (nproma,nlev,nblks_e)
  UnmanagedConstT3D p_edge_in_view(p_edge_in, nproma, nlev, nblks_e);

  // line indices of edges of triangles, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iidx_view(edge_idx, nproma, nblks_c, 3); // edge_idx_view

  // block indices of edges of triangles, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iblk_view(edge_blk, nproma, nblks_c, 3); // edge_blk_view

  // coefficients for (area weighted) interpolation, dim:
  // (nproma,3-cell_type,nblks_c)
  UnmanagedConstT3D coeff_int_view(coeff_int, nproma, 3, nblks_c);

  // cell based scalar output field, dim: (nproma,nlev,nblks_c)
  UnmanagedT3D p_cell_out_view(p_cell_out, nproma, nlev, nblks_c);

  int i_startidx, i_endidx;

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "edges2cells_scalar_lib_inner", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          p_cell_out_view(jc, jk, jb) =
              coeff_int_view(jc, 0, jb) * p_edge_in_view(iidx_view(jc, jb, 0),
                                                         jk,
                                                         iblk_view(jc, jb, 0)) +
              coeff_int_view(jc, 1, jb) * p_edge_in_view(iidx_view(jc, jb, 1),
                                                         jk,
                                                         iblk_view(jc, jb, 1)) +
              coeff_int_view(jc, 2, jb) * p_edge_in_view(iidx_view(jc, jb, 2),
                                                         jk,
                                                         iblk_view(jc, jb, 2));
        });
    Kokkos::fence();
  }
}

//------------------------------------------------------------------------
//>
///  Computes  average of scalar fields from centers of cells to vertices.
///
template <typename T, typename S>
void cells2verts_scalar_lib(const T *p_cell_in, const int *vert_cell_idx,
                            const int *vert_cell_blk, const S *coeff_int,
                            S *p_vert_out, const int i_startblk,
                            const int i_endblk, const int i_startidx_in,
                            const int i_endidx_in, const int slev,
                            const int elev, const int nproma, const int nlev,
                            const int nblks_c, const int nblks_v,
                            const bool lacc, const bool acc_async) {
  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<const S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstS3D;
  typedef Kokkos::View<S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedS3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  // cell based scalar input field, dim: (nproma,nlev,nblks_c)
  UnmanagedConstT3D p_cell_in_view(p_cell_in, nproma, nlev, nblks_c);

  // line indices of cells around each vertex, dim: (nproma,nblks_v, 6)
  UnmanagedConstInt3D iidx_view(vert_cell_idx, nproma, nblks_v,
                                6); // vert_cell_idx_view

  // block indices of cells around each vertex, dim: (nproma,nblks_v, 6)
  UnmanagedConstInt3D iblk_view(vert_cell_blk, nproma, nblks_v,
                                6); // vert_cell_blk_view

  // coefficients for interpolation, dim: (nproma,9-cell_type,nblks_v)
  UnmanagedConstS3D coeff_int_view(coeff_int, nproma, 6, nblks_v);

  // vertex based scalar output field, dim: (nproma,nlev,nblks_c)
  UnmanagedS3D p_vert_out_view(p_vert_out, nproma, nlev, nblks_c);

  int i_startidx, i_endidx;

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

    get_indices_v_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "cells2verts_scalar_lib", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jv) {
          p_vert_out_view(jv, jk, jb) =
              coeff_int_view(jv, 0, jb) * p_cell_in_view(iidx_view(jv, jb, 0),
                                                         jk,
                                                         iblk_view(jv, jb, 0)) +
              coeff_int_view(jv, 1, jb) * p_cell_in_view(iidx_view(jv, jb, 1),
                                                         jk,
                                                         iblk_view(jv, jb, 1)) +
              coeff_int_view(jv, 2, jb) * p_cell_in_view(iidx_view(jv, jb, 2),
                                                         jk,
                                                         iblk_view(jv, jb, 2)) +
              coeff_int_view(jv, 3, jb) * p_cell_in_view(iidx_view(jv, jb, 3),
                                                         jk,
                                                         iblk_view(jv, jb, 3)) +
              coeff_int_view(jv, 4, jb) * p_cell_in_view(iidx_view(jv, jb, 4),
                                                         jk,
                                                         iblk_view(jv, jb, 4)) +
              coeff_int_view(jv, 5, jb) * p_cell_in_view(iidx_view(jv, jb, 5),
                                                         jk,
                                                         iblk_view(jv, jb, 5));
        });
    Kokkos::fence();
  }
}

//-------------------------------------------------------------------------
//>
///  Same as above, but provides output optionally in single precision and
///  assumes reversed index order of the output field in loop exchange mode
///
template <typename T, typename S>
void cells2verts_scalar_ri_lib(const T *p_cell_in, const int *vert_cell_idx,
                               const int *vert_cell_blk, const T *coeff_int,
                               S *p_vert_out, const int i_startblk,
                               const int i_endblk, const int i_startidx_in,
                               const int i_endidx_in, const int slev,
                               const int elev, const int nproma, const int nlev,
                               const int nblks_c, const int nblks_v,
                               const bool lacc, const bool acc_async) {
  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<S ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedS3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  // cell based scalar input field, dim: (nproma,nlev,nblks_c)
  UnmanagedConstT3D p_cell_in_view(p_cell_in, nproma, nlev, nblks_c);

  // line indices of cells around each vertex, dim: (nproma,nblks_v, 6)
  UnmanagedConstInt3D iidx_view(vert_cell_idx, nproma, nblks_v,
                                6); // vert_cell_idx_view

  // block indices of cells around each vertex, dim: (nproma,nblks_v, 6)
  UnmanagedConstInt3D iblk_view(vert_cell_blk, nproma, nblks_v,
                                6); // vert_cell_blk_view

  // coefficients for interpolation, dim: (nproma,9-cell_type,nblks_v)
  UnmanagedConstT3D coeff_int_view(coeff_int, nproma, 6, nblks_v);

  // vertex based scalar output field, dim: (nproma,nlev,nblks_c)
#ifdef __LOOP_EXCHANGE
  UnmanagedS3D p_vert_out_view(p_vert_out, nproma, nlev, nblks_c);
#else
  UnmanagedS3D p_vert_out_view(p_vert_out, nlev, nproma, nblks_c);
#endif

  int i_startidx, i_endidx;

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {

    get_indices_v_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "cells2verts_scalar_ri_lib", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jv) {

#ifdef __LOOP_EXCHANGE
          p_vert_out_view(jv, jk, jb) =
#else
          p_vert_out_view(jk, jv, jb) =
#endif
              coeff_int_view(jv, 0, jb) * p_cell_in_view(iidx_view(jv, jb, 0),
                                                         jk,
                                                         iblk_view(jv, jb, 0)) +
              coeff_int_view(jv, 1, jb) * p_cell_in_view(iidx_view(jv, jb, 1),
                                                         jk,
                                                         iblk_view(jv, jb, 1)) +
              coeff_int_view(jv, 2, jb) * p_cell_in_view(iidx_view(jv, jb, 2),
                                                         jk,
                                                         iblk_view(jv, jb, 2)) +
              coeff_int_view(jv, 3, jb) * p_cell_in_view(iidx_view(jv, jb, 3),
                                                         jk,
                                                         iblk_view(jv, jb, 3)) +
              coeff_int_view(jv, 4, jb) * p_cell_in_view(iidx_view(jv, jb, 4),
                                                         jk,
                                                         iblk_view(jv, jb, 4)) +
              coeff_int_view(jv, 5, jb) * p_cell_in_view(iidx_view(jv, jb, 5),
                                                         jk,
                                                         iblk_view(jv, jb, 5));
        });
    Kokkos::fence();
  }
}

//-------------------------------------------------------------------------
//>
///  Computes  average of scalar fields from vertices to centers of cells.
///
template <typename T>
void verts2cells_scalar_lib(const T *p_vert_in, const int *cell_index_idx,
                            const int *cell_vertex_blk, const T *coeff_int,
                            T *p_cell_out, const int nblks_c,
                            const int npromz_c, const int slev, const int elev,
                            const int nproma, const int nlev, const int nblks_v,
                            const bool lacc) {
  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  // cell based scalar input field, dim: (nproma,nlev,nblks_v)
  UnmanagedConstT3D p_vert_in_view(p_vert_in, nproma, nlev, nblks_v);

  // line indices of vertices of triangles, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iidx_view(cell_index_idx, nproma, nblks_c,
                                3); // cell_vertex_idx

  // block indices of vertices of triangles, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iblk_view(cell_vertex_blk, nproma, nblks_c,
                                3); // cell_vertex_blk

  // coefficients for interpolation, dim: (nproma, 3, nblks_c)
  UnmanagedConstT3D coeff_int_view(coeff_int, nproma, 3, nblks_c);

  // vertex based scalar output field, dim: (nproma,nlev,nblks_c)
  UnmanagedT3D p_cell_out_view(p_cell_out, nproma, nlev, nblks_c);

  for (int jb = 0; jb < nblks_c; ++jb) {

    int nlen;
    if (jb != nblks_c) {
      nlen = nproma;
    } else {
      nlen = npromz_c;
    }

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy({slev, 0},
                                                       {elev + 1, nlen});

    Kokkos::parallel_for(
        "cell_avg_lib_inner", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          p_cell_out_view(jc, jk, jb) =
              coeff_int_view(jc, 0, jb) * p_vert_in_view(iidx_view(jc, jb, 0),
                                                         jk,
                                                         iblk_view(jc, jb, 0)) +
              coeff_int_view(jc, 1, jb) * p_vert_in_view(iidx_view(jc, jb, 1),
                                                         jk,
                                                         iblk_view(jc, jb, 1)) +
              coeff_int_view(jc, 2, jb) * p_vert_in_view(iidx_view(jc, jb, 2),
                                                         jk,
                                                         iblk_view(jc, jb, 2));
        });
    Kokkos::fence();
  }
}

//-------------------------------------------------------------------------
//>
/// Computes the average of a cell-based variable.
///
/// Computes the average of a cell-based variable
/// over its original location and the neighboring triangles.
/// Version with variable weighting coefficients, computed such that
/// linear horizontal gradients are not aliased into a checkerboard noise
/// input:  lives on centers of triangles
/// output: lives on centers of triangles
///
template <typename T>
void cell_avg_lib(const T *psi_c, const int *cell_neighbor_idx,
                  const int *cell_neighbor_blk, const T *avg_coeff,
                  T *avg_psi_c, const int i_startblk, const int i_endblk,
                  const int i_startidx_in, const int i_endidx_in,
                  const int slev, const int elev, const int nproma,
                  const int nlev, const int nblks_c, const bool lacc) {
  // Wrap raw pointers in unmanaged Kokkos Views.
  typedef Kokkos::View<const T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedConstT3D;
  typedef Kokkos::View<T ***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged>
      UnmanagedT3D;
  typedef Kokkos::View<const int ***, Kokkos::LayoutLeft,
                       Kokkos::MemoryUnmanaged>
      UnmanagedConstInt3D;

  // cell based variable before averaging, dim: (nproma,nlev,nblks_c)
  UnmanagedConstT3D psi_c_view(psi_c, nproma, nlev, nblks_c);
  // line indices of triangles next to each cell, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iidx_view(cell_neighbor_idx, nproma, nblks_c,
                                3); // cell_neighbour_idx
  // block indices of triangles next to each cell, dim: (nproma,nblks_c, 3)
  UnmanagedConstInt3D iblk_view(cell_neighbor_blk, nproma, nblks_c,
                                3); // cell_neighbour_blk
  // averaging coefficients, dim: (nproma,nlev,nblks_c)
  UnmanagedConstT3D avg_coeff_view(avg_coeff, nproma, nlev, nblks_c);

  // cell based variable after averaging, dim: (nproma,nlev,nblks_c)
  UnmanagedT3D avg_psi_c_view(avg_psi_c, nproma, nlev, nblks_c);

  int i_startidx, i_endidx;

  for (int jb = i_startblk; jb < i_endblk + 1; ++jb) {
    get_indices_c_lib(i_startidx_in, i_endidx_in, nproma, jb, i_startblk,
                      i_endblk, i_startidx, i_endidx);

    Kokkos::MDRangePolicy<Kokkos::Rank<2>> innerPolicy(
        {slev, i_startidx}, {elev + 1, i_endidx + 1});

    Kokkos::parallel_for(
        "cell_avg_lib_inner", innerPolicy,
        KOKKOS_LAMBDA(const int jk, const int jc) {
          //  calculate the weighted average

          avg_psi_c_view(jc, jk, jb) =
              psi_c_view(jc, jk, jb) * avg_coeff_view(jc, 0, jb) +
              psi_c_view(iidx_view(jc, jb, 0), jk, iblk_view(jc, jb, 0)) *
                  avg_coeff_view(jc, 1, jb) +
              psi_c_view(iidx_view(jc, jb, 1), jk, iblk_view(jc, jb, 1)) *
                  avg_coeff_view(jc, 2, jb) +
              psi_c_view(iidx_view(jc, jb, 2), jk, iblk_view(jc, jb, 2)) *
                  avg_coeff_view(jc, 3, jb);
        });
    Kokkos::fence();
  }
}

//-----------------------------------------------------------------------
//
//  Explicit Instantiations
//
//-----------------------------------------------------------------------

template void verts2edges_scalar_lib<double>(
    const double *p_vertex_in, const int *edge_vertex_idx,
    const int *edge_vertex_blk, const double *coeff_int, double *p_edge_out,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const int nlev, const int nblks_v, const int nblks_e, const bool lacc);

template void verts2edges_scalar_lib<float>(
    const float *p_vertex_in, const int *edge_vertex_idx,
    const int *edge_vertex_blk, const float *coeff_int, float *p_edge_out,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const int nlev, const int nblks_v, const int nblks_e, const bool lacc);

template void cells2edges_scalar_lib<double, double>(
    const double *p_cell_in, const int *edge_cell_idx, const int *edge_cell_blk,
    const double *coeff_int, double *p_edge_out, const int *i_startblk_in,
    const int *i_endblk_in, const int *i_startidx_in, const int *i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblk_c, const int nblks_e, const int patch_id,
    const bool l_limited_area, const bool lfill_latbc, const bool lacc);

template void cells2edges_scalar_lib<float, float>(
    const float *p_cell_in, const int *edge_cell_idx, const int *edge_cell_blk,
    const float *coeff_int, float *p_edge_out, const int *i_startblk_in,
    const int *i_endblk_in, const int *i_startidx_in, const int *i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblk_c, const int nblks_e, const int patch_id,
    const bool l_limited_area, const bool lfill_latbc, const bool lacc);

// sp2dp
template void cells2edges_scalar_lib<float, double>(
    const float *p_cell_in, const int *edge_cell_idx, const int *edge_cell_blk,
    const double *coeff_int, double *p_edge_out, const int *i_startblk_in,
    const int *i_endblk_in, const int *i_startidx_in, const int *i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblk_c, const int nblks_e, const int patch_id,
    const bool l_limited_area, const bool lfill_latbc, const bool lacc);

template void edges2verts_scalar_lib<double>(
    const double *p_edge_in, const int *vert_edge_idx, const int *vert_edge_blk,
    const double *v_int, double *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_e, const int nblks_v, const bool lacc);

template void edges2verts_scalar_lib<float>(
    const float *p_edge_in, const int *vert_edge_idx, const int *vert_edge_blk,
    const float *v_int, float *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_e, const int nblks_v, const bool lacc);

template void edges2cells_scalar_lib<double>(
    const double *p_edge_in, const int *edge_idx, const int *edge_blk,
    const double *coeff_int, double *p_cell_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_e, const int nblks_c, const bool lacc);

template void edges2cells_scalar_lib<float>(
    const float *p_edge_in, const int *edge_idx, const int *edge_blk,
    const float *coeff_int, float *p_cell_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_e, const int nblks_c, const bool lacc);

template void cells2verts_scalar_lib<double, double>(
    const double *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const double *coeff_int, double *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void cells2verts_scalar_lib<float, double>(
    const float *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const double *coeff_int, double *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void cells2verts_scalar_lib<float, float>(
    const float *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const float *coeff_int, float *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void cells2verts_scalar_ri_lib<double, double>(
    const double *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const double *coeff_int, double *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void cells2verts_scalar_ri_lib<double, float>(
    const double *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const double *coeff_int, float *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void cells2verts_scalar_ri_lib<float, float>(
    const float *p_cell_in, const int *vert_cell_idx, const int *vert_cell_blk,
    const float *coeff_int, float *p_vert_out, const int i_startblk,
    const int i_endblk, const int i_startidx_in, const int i_endidx_in,
    const int slev, const int elev, const int nproma, const int nlev,
    const int nblks_c, const int nblks_v, const bool lacc,
    const bool acc_async);

template void verts2cells_scalar_lib<double>(
    const double *p_vert_in, const int *cell_index_idx,
    const int *cell_vertex_blk, const double *coeff_int, double *p_cell_out,
    const int nblks_c, const int npromz_c, const int slev, const int elev,
    const int nproma, const int nlev, const int nblks_v, const bool lacc);

template void verts2cells_scalar_lib<float>(
    const float *p_vert_in, const int *cell_index_idx,
    const int *cell_vertex_blk, const float *coeff_int, float *p_cell_out,
    const int nblks_c, const int npromz_c, const int slev, const int elev,
    const int nproma, const int nlev, const int nblks_v, const bool lacc);

template void cell_avg_lib<double>(
    const double *psi_c, const int *cell_neighbor_idx,
    const int *cell_neighbor_blk, const double *avg_coeff, double *avg_psi_c,
    const int i_startblk, const int i_endblk, const int i_startidx_in,
    const int i_endidx_in, const int slev, const int elev, const int nproma,
    const int nlev, const int nblks_c, const bool lacc);

template void
cell_avg_lib<float>(const float *psi_c, const int *cell_neighbor_idx,
                    const int *cell_neighbor_blk, const float *avg_coeff,
                    float *avg_psi_c, const int i_startblk, const int i_endblk,
                    const int i_startidx_in, const int i_endidx_in,
                    const int slev, const int elev, const int nproma,
                    const int nlev, const int nblks_c, const bool lacc);
