// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_lib_interpolation_scalar.hpp"
#include <Kokkos_Core.hpp>
#include <gtest/gtest.h>
#include <vector>

// Free-function helpers for 3D and 4D array sizes (assumed column-major)
template <typename T> size_t num_elements_3d(int d1, int d2, int d3) {
  return static_cast<size_t>(d1) * d2 * d3;
}

template <typename T> size_t num_elements_4d(int d1, int d2, int d3, int d4) {
  return static_cast<size_t>(d1) * d2 * d3 * d4;
}

// Define a helper struct that holds the two types.
template <typename InT, typename OutT> struct MixedPrecision {
  using in_type = InT;
  using out_type = OutT;
};

// Define the list of type pairs we want to test.
typedef ::testing::Types<MixedPrecision<double, double>,
                         MixedPrecision<double, float>,
                         MixedPrecision<float, float>>
    MixedTypes;

typedef ::testing::Types<MixedPrecision<double, double>,
                         MixedPrecision<float, double>,
                         MixedPrecision<float, float>>
    MixedTypesSP2DP;

// Shared dimensions for all routines and classes
class interp_dimensions {
public:
  // Constant dimensions.
  static constexpr int nproma = 16; // inner loop length
  static constexpr int nlev = 7;    // number of vertical levels
  static constexpr int nblks_c = 2; // number of cell blocks
  static constexpr int nblks_e = 2; // number of edge blocks (for p_e_in)
  static constexpr int nblks_v = 2; // number of vertex blocks

  // Parameter values.
  const int i_startblk = 0;
  const int i_endblk = 1; // Test blocks [0, 1]
  const int i_startidx = 2;
  const int i_endidx = nproma - 3; // Partial range: 2 .. nproma-3
  const int slev = 1;
  const int elev = nlev - 1;    // Partial vertical range (1 .. nlev-1)
  const bool lacc = false;      // Not using ACC-specific behavior.
  const bool acc_async = false; // No asynchronous execution.
};

template <typename T>
class InterpolationScalarTypedTestFixture : public ::testing::Test,
                                            public interp_dimensions {
public:
  // Arrays used for verts2edges
  std::vector<T> p_vertex_in;       // Dimensions: (nproma, nlev, nblks_v)
  std::vector<int> edge_vertex_idx; // Dimensions: (nproma, nblks_e, 4)
  std::vector<int> edge_vertex_blk; // Dimensions: (nproma, nblks_e, 4)
  std::vector<T> coeff_int_edges;   // Dimensions: (nproma, 2, nblks_e)
  std::vector<T> p_edge_out;        // Dimensions: (nproma, nlev, nblks_e)

  // Arrays used for edges2verts
  std::vector<T> p_edge_in;       // Dimensions: (nproma, nlev, nblks_e)
  std::vector<int> edge_vert_idx; // Dimensions: (nproma, nblks_e, 6)
  std::vector<int> edge_vert_blk; // Dimensions: (nproma, nblks_e, 6)
  std::vector<T> v_int;           // Dimensions: (nproma, 6, nblks_v)
  std::vector<T> p_vert_out;      // Dimensions: (nproma, nlev, nblks_v)

  // Arrays used for edges2cells
  // std::vector<T> p_edge_in;        // Dimensions: (nproma, nlev, nblks_e)
  std::vector<int> edge_idx;      // Dimensions: (nproma, nblks_c, 3)
  std::vector<int> edge_blk;      // Dimensions: (nproma, nblks_c, 3)
  std::vector<T> coeff_int_cells; // Dimensions: (nproma, 3, nblks_c)
  std::vector<T> p_cell_out;      // Dimensions: (nproma, nlev, nblks_c)

  // Arrays used for verts2cells
  std::vector<T> p_vert_in;        // Dimensions: (nproma, nlev, nblks_v)
  std::vector<int> cell_index_idx; // Dimensions: (nproma, nblks_c, 3)
  std::vector<int> cell_index_blk; // Dimensions: (nproma, nblks_c, 3)

  // Arrays used for avg_lib
  std::vector<T> psi_c;               // Dimensions: (nproma, nlev, nblks_c)
  std::vector<int> cell_neighbor_idx; // Dimensions: (nproma, nblks_c, 3)
  std::vector<int> cell_neighbor_blk; // Dimensions: (nproma, nblks_c, 3)
  std::vector<T> avg_coeff;           // Dimensions: (nproma, nlev, nblks_c)
  std::vector<T> avg_psi_c;           // Dimensions: (nproma, nlev, nblks_c)

  const int cell_type = 6;
  const int npromz_c = 32;

  InterpolationScalarTypedTestFixture() {
    // Allocate and initialize arrays needed for verts2edges
    p_vertex_in.resize(num_elements_3d<T>(nproma, nlev, nblks_v),
                       static_cast<T>(1));
    edge_vertex_idx.resize(num_elements_3d<int>(nproma, nblks_e, 4), 1);
    edge_vertex_blk.resize(num_elements_3d<int>(nproma, nblks_e, 4), 0);
    coeff_int_edges.resize(num_elements_3d<T>(nproma, 2, nblks_e),
                           static_cast<T>(1));

    p_edge_out.resize(num_elements_3d<T>(nproma, nlev, nblks_e),
                      static_cast<T>(0));

    // Allocate & Initialize arrays needed for edges2verts
    p_edge_in.resize(num_elements_3d<T>(nproma, nlev, nblks_e),
                     static_cast<T>(1));
    edge_vert_idx.resize(num_elements_3d<int>(nproma, nblks_e, 6), 1);
    edge_vert_blk.resize(num_elements_3d<int>(nproma, nblks_e, 6), 0);
    v_int.resize(num_elements_3d<T>(nproma, 6, nblks_v), static_cast<T>(1));

    p_vert_out.resize(num_elements_3d<T>(nproma, nlev, nblks_v),
                      static_cast<T>(0));

    // Allocate & Initialize arrays needed for edges2cells
    edge_idx.resize(num_elements_3d<int>(nproma, nblks_c, 3), 1);
    edge_blk.resize(num_elements_3d<int>(nproma, nblks_c, 3), 0);
    coeff_int_cells.resize(num_elements_3d<T>(nproma, 3, nblks_c),
                           static_cast<T>(1));

    p_cell_out.resize(num_elements_3d<T>(nproma, nlev, nblks_c),
                      static_cast<T>(0));

    // Allocate and initialize arrays needed for verts2cells
    p_vert_in.resize(num_elements_3d<T>(nproma, nlev, nblks_v),
                     static_cast<T>(1));
    cell_index_idx.resize(num_elements_3d<int>(nproma, nblks_c, 3), 1);
    cell_index_blk.resize(num_elements_3d<int>(nproma, nblks_c, 3), 0);

    // Allocate and initialize arrays needed for avg_lib
    psi_c.resize(num_elements_3d<T>(nproma, nlev, nblks_c), static_cast<T>(1));
    cell_neighbor_idx.resize(num_elements_3d<int>(nproma, nblks_c, 3), 1);
    cell_neighbor_blk.resize(num_elements_3d<int>(nproma, nblks_c, 3), 0);
    avg_coeff.resize(num_elements_3d<T>(nproma, nlev, nblks_c),
                     static_cast<T>(1));

    // Allocate output arrays and initialize to zero.
    avg_psi_c.resize(num_elements_3d<T>(nproma, nlev, nblks_c),
                     static_cast<T>(0));
  }
};

typedef ::testing::Types<float, double> SingleType;

TYPED_TEST_SUITE(InterpolationScalarTypedTestFixture, SingleType);

////////////////////////////////////////////////////////////////////////////////
//
// ! verts2edges
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarTypedTestFixture, Verts2Edges) {

  verts2edges_scalar_lib<TypeParam>(
      this->p_vertex_in.data(), this->edge_vertex_idx.data(),
      this->edge_vertex_blk.data(), this->coeff_int_edges.data(),
      this->p_edge_out.data(), this->i_startblk, this->i_endblk,
      this->i_startidx, this->i_endidx, this->slev, this->elev, this->nproma,
      this->nlev, this->nblks_v, this->nblks_e, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx, i_endidx] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 2 stencil points,
        // expect 2.
        EXPECT_NEAR(this->p_edge_out[idx], static_cast<TypeParam>(2),
                    static_cast<TypeParam>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// ! edges2verts
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarTypedTestFixture, Edges2Verts) {

  edges2verts_scalar_lib<TypeParam>(
      this->p_edge_in.data(), this->edge_vert_idx.data(),
      this->edge_vert_blk.data(), this->v_int.data(), this->p_vert_out.data(),
      this->i_startblk, this->i_endblk, this->i_startidx, this->i_endidx,
      this->slev, this->elev, this->nproma, this->nlev, this->nblks_e,
      this->nblks_v, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 6 stencil points,
        // expect 6.
        EXPECT_NEAR(this->p_vert_out[idx], static_cast<TypeParam>(6),
                    static_cast<TypeParam>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// ! edges2cells
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarTypedTestFixture, Edges2Cells) {

  edges2cells_scalar_lib<TypeParam>(
      this->p_edge_in.data(), this->edge_idx.data(), this->edge_blk.data(),
      this->coeff_int_cells.data(), this->p_cell_out.data(), this->i_startblk,
      this->i_endblk, this->i_startidx, this->i_endidx, this->slev, this->elev,
      this->nproma, this->nlev, this->nblks_e, this->nblks_c, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 3 stencil points,
        // expect 3.
        EXPECT_NEAR(this->p_cell_out[idx], static_cast<TypeParam>(3),
                    static_cast<TypeParam>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

TYPED_TEST(InterpolationScalarTypedTestFixture, Verts2Cells) {

  verts2cells_scalar_lib<TypeParam>(
      this->p_vert_in.data(), this->cell_index_idx.data(),
      this->cell_index_blk.data(), this->coeff_int_cells.data(),
      this->p_cell_out.data(), this->nblks_c, this->npromz_c, this->slev,
      this->elev, this->nproma, this->nlev, this->nblks_v, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 3 stencil points,
        // expect 3.
        EXPECT_NEAR(this->p_cell_out[idx], static_cast<TypeParam>(3),
                    static_cast<TypeParam>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// ! cell_avg
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarTypedTestFixture, AvgLib) {

  // Call the function
  cell_avg_lib<TypeParam>(this->psi_c.data(), this->cell_neighbor_idx.data(),
                          this->cell_neighbor_blk.data(),
                          this->avg_coeff.data(), this->avg_psi_c.data(),
                          this->i_startblk, this->i_endblk, this->i_startidx,
                          this->i_endidx, this->slev, this->elev, this->nproma,
                          this->nlev, this->nblks_c, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 4 stencil points,
        // expect 4.
        EXPECT_NEAR(this->avg_psi_c[idx], static_cast<TypeParam>(4),
                    static_cast<TypeParam>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

template <typename TypePair>
class InterpolationScalarMixedTestFixture : public ::testing::Test,
                                            public interp_dimensions {
public:
  using InType = typename TypePair::in_type;
  using OutType = typename TypePair::out_type;

  // Arrays used for cells2edges
  std::vector<InType> p_cell_in;        // Dimensions: (nproma, nlev, nblks_c)
  std::vector<int> edge_cell_idx;       // Dimensions: (nproma, nblks_e, 2)
  std::vector<int> edge_cell_blk;       // Dimensions: (nproma, nblks_e, 2)
  std::vector<OutType> coeff_int_edges; // Dimensions: (nproma, 2, nblks_e)
  std::vector<OutType> p_edge_out;      // Dimensions: (nproma, nlev, nblks_e)

  // Further parameters for cells2edges
  const int patch_id = 0;
  const bool l_limited_area = false;
  const bool lfill_latbc = false;
  std::vector<int> i_startblk_in; // Dimensions: (2)
  std::vector<int> i_endblk_in;   // Dimensions: (2)
  std::vector<int> i_startidx_in; // Dimensions: (2)
  std::vector<int> i_endidx_in;   // Dimensions: (2)

  // Arrays used for cells2verts
  std::vector<int> vert_cell_idx;       // Dimensions: (nproma, nblks_v, 6)
  std::vector<int> vert_cell_blk;       // Dimensions: (nproma, nblks_v, 6)
  std::vector<OutType> coeff_int_verts; // Dimensions: (nproma, 6, nblks_v)
  std::vector<OutType> p_vert_out;      // Dimensions: (nproma, nlev, nblks_v)

  InterpolationScalarMixedTestFixture() {
    // Allocate and initialize arrays needed for cells2edges
    p_cell_in.resize(num_elements_3d<InType>(nproma, nlev, nblks_c),
                     static_cast<InType>(1));
    edge_cell_idx.resize(num_elements_3d<int>(nproma, nblks_e, 2), 1);
    edge_cell_blk.resize(num_elements_3d<int>(nproma, nblks_e, 2), 0);
    coeff_int_edges.resize(num_elements_3d<InType>(nproma, 2, nblks_e),
                           static_cast<OutType>(1));

    p_edge_out.resize(num_elements_3d<OutType>(nproma, nlev, nblks_e),
                      static_cast<OutType>(0));

    // Allocate neighbour indexes for cells2edges
    i_startblk_in.resize(2, i_startblk);
    i_endblk_in.resize(2, i_endblk);
    i_startidx_in.resize(2, i_startidx);
    i_endidx_in.resize(2, i_endidx);

    // Allocate & Initialize arrays needed for cells2verts
    vert_cell_idx.resize(num_elements_3d<int>(nproma, nblks_v, 6), 1);
    vert_cell_blk.resize(num_elements_3d<int>(nproma, nblks_v, 6), 0);
    coeff_int_verts.resize(num_elements_3d<InType>(nproma, 6, nblks_v),
                           static_cast<OutType>(1));

    p_vert_out.resize(num_elements_3d<OutType>(nproma, nlev, nblks_v),
                      static_cast<OutType>(0));
  }
};

TYPED_TEST_SUITE(InterpolationScalarMixedTestFixture, MixedTypesSP2DP);

////////////////////////////////////////////////////////////////////////////////
//
// ! cells2edges
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarMixedTestFixture, Cells2Edges) {
  using InType = typename TestFixture::InType;
  using OutType = typename TestFixture::OutType;

  // Call the function
  cells2edges_scalar_lib<InType, OutType>(
      this->p_cell_in.data(), this->edge_cell_idx.data(),
      this->edge_cell_blk.data(), this->coeff_int_edges.data(),
      this->p_edge_out.data(), this->i_startblk_in.data(),
      this->i_endblk_in.data(), this->i_startidx_in.data(),
      this->i_endidx_in.data(), this->slev, this->elev, this->nproma,
      this->nlev, this->nblks_c, this->nblks_e, this->patch_id,
      this->l_limited_area, this->lfill_latbc, this->lacc);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 2 stencil points,
        // expect 2.
        EXPECT_NEAR(this->p_edge_out[idx], static_cast<OutType>(2),
                    static_cast<OutType>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// ! cells2verts
//
////////////////////////////////////////////////////////////////////////////////

TYPED_TEST(InterpolationScalarMixedTestFixture, Cells2Verts) {
  using InType = typename TestFixture::InType;
  using OutType = typename TestFixture::OutType;

  cells2verts_scalar_lib<InType, OutType>(
      this->p_cell_in.data(), this->vert_cell_idx.data(),
      this->vert_cell_blk.data(), this->coeff_int_verts.data(),
      this->p_vert_out.data(), this->i_startblk, this->i_endblk,
      this->i_startidx, this->i_endidx, this->slev, this->elev, this->nproma,
      this->nlev, this->nblks_c, this->nblks_v, this->lacc, this->acc_async);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 6 stencil points,
        // expect 6.
        EXPECT_NEAR(this->p_vert_out[idx], static_cast<OutType>(6),
                    static_cast<OutType>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//
// ! cells2verts ri
//
////////////////////////////////////////////////////////////////////////////////

// The test for cells2verts_ri is similar to cells2verts, but is done here
// separtely to avoid as a differebt template instantiation is needed for the
// function call
template <typename Types>
class Cells2vertsriScalarLibTestFixture : public testing::Test,
                                          public interp_dimensions {
public:
  using InType = typename Types::in_type;
  using OutType = typename Types::out_type;

  // Arrays stored in std::vector.
  std::vector<InType> p_cell_in;   // Dimensions: (nproma, nlev, nblks_c)
  std::vector<int> vert_cell_idx;  // Dimensions: (nproma, nblks_v, 6)
  std::vector<int> vert_cell_blk;  // Dimensions: (nproma, nblks_v, 6)
  std::vector<InType> coeff_int;   // Dimensions: (nproma, 6, nblks_v)
  std::vector<OutType> p_vert_out; // Dimensions: (nproma, nlev, nblks_v)

  Cells2vertsriScalarLibTestFixture() {
    // Allocate and initialize inputs.
    p_cell_in.resize(num_elements_3d<InType>(nproma, nlev, nblks_c),
                     static_cast<InType>(1));
    vert_cell_idx.resize(num_elements_3d<int>(nproma, nblks_v, 6), 1);
    vert_cell_blk.resize(num_elements_3d<int>(nproma, nblks_v, 6), 0);
    coeff_int.resize(num_elements_3d<InType>(nproma, 6, nblks_v),
                     static_cast<InType>(1));

    // Allocate output arrays and initialize to zero.
    p_vert_out.resize(num_elements_3d<OutType>(nproma, nlev, nblks_v),
                      static_cast<OutType>(0));
  }
};

// Add test suite
TYPED_TEST_SUITE(Cells2vertsriScalarLibTestFixture, MixedTypes);

// Add test
TYPED_TEST(Cells2vertsriScalarLibTestFixture, Cells2VertsRI) {
  using InType = typename TestFixture::InType;
  using OutType = typename TestFixture::OutType;

  // Call the function
  cells2verts_scalar_ri_lib<InType, OutType>(
      this->p_cell_in.data(), this->vert_cell_idx.data(),
      this->vert_cell_blk.data(), this->coeff_int.data(),
      this->p_vert_out.data(), this->i_startblk, this->i_endblk,
      this->i_startidx, this->i_endidx, this->slev, this->elev, this->nproma,
      this->nlev, this->nblks_c, this->nblks_v, this->lacc, this->acc_async);

  // Check the outputs only for blocks in the range
  // { [i_startblk, i_endblk], [slev,elev], [i_startidx_in, i_endidx_in] }
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = this->slev; level < this->elev; ++level) {
      for (int i = this->i_startidx; i < this->i_endidx; ++i) {
        // Compute the linear index for a 3D array in column-major order:
#ifdef __LOOP_EXCHANGE
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
#else
        size_t idx = level + i * this->nlev + block * this->nproma * this->nlev;
#endif
        // Since every contribution is 1 and there are 6 stencil points,
        // expect 6.
        EXPECT_NEAR(this->p_vert_out[idx], static_cast<OutType>(6),
                    static_cast<OutType>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}
